/**
 * 
 * Copyright 2004 Hiram Chirino
 * Copyright 2004 Protique Ltd
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License. 
 * 
 **/
package org.codehaus.activemq.store.jdbc;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import java.sql.Connection;
import java.util.LinkedList;

/**
 * Helps keep track of the current transaction/JDBC connection.
 *
 * @version $Revision: 1.1 $
 */
public class TransactionContext {
    private static final Log log = LogFactory.getLog(TransactionContext.class);
    private static ThreadLocal threadLocalTxn = new ThreadLocal();

    /**
     * @return the current thread local connection that is associated
     *  with the JMS transaction or null if there is no 
     *  transaction in progress.
     */
    public static Connection getConnection() {
        LinkedList list = (LinkedList) threadLocalTxn.get();
        if (list != null && !list.isEmpty()) {
            return (Connection) list.getFirst();
        }
        throw new IllegalStateException("Attempt to get a connection when no transaction in progress");
    }

    /**
     * Pops off the current Connection from the stack
     */
    public static Connection popConnection() {
        LinkedList list = (LinkedList) threadLocalTxn.get();
        if (list == null || list.isEmpty()) {
            log.warn("Attempt to pop connection when no transaction in progress");
            return null;
        }
        else {
            return (Connection) list.removeFirst();
        }
    }

    /**
     * Sets the current transaction, possibly including nesting
     */
    public static void pushConnection(Connection connection) {
        LinkedList list = (LinkedList) threadLocalTxn.get();
        if (list == null) {
            list = new LinkedList();
            threadLocalTxn.set(list);
        }
        list.addLast(connection);
    }

    public static int getTransactionCount() {
        LinkedList list = (LinkedList) threadLocalTxn.get();
        if (list != null) {
            return list.size();
        }
        return 0;
    }

}
