/**
 * 
 * Copyright 2004 Protique Ltd
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License. 
 * 
 **/

package org.codehaus.activemq.transport.tcp;

import EDU.oswego.cs.dl.util.concurrent.BoundedBuffer;
import EDU.oswego.cs.dl.util.concurrent.PooledExecutor;
import EDU.oswego.cs.dl.util.concurrent.SynchronizedBoolean;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.codehaus.activemq.message.WireFormat;
import org.codehaus.activemq.transport.TransportServerChannelSupport;

import javax.jms.JMSException;
import java.io.IOException;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.URI;
import java.net.UnknownHostException;

/**
 * Binds to a well known port and listens for Sockets ...
 *
 * @version $Revision: 1.17 $
 */
public class TcpTransportServerChannel extends TransportServerChannelSupport implements Runnable {
    private static final Log log = LogFactory.getLog(TcpTransportServerChannel.class);
    protected static final int DEFEAULT_BACKLOG = 500;

    private WireFormat wireFormat;
    protected String bindAddressURI;
    private Thread serverSocketThread;
    private ServerSocket serverSocket;
    private SynchronizedBoolean closed;
    private SynchronizedBoolean started;

    private boolean useAsyncSend = false;
    private int maxOutstandingMessages = 10;
    private int backlog = DEFEAULT_BACKLOG;

    /**
     * Default Constructor
     *
     * @param bindAddr
     * @throws JMSException
     */
    public TcpTransportServerChannel(WireFormat wireFormat, URI bindAddr) throws JMSException {
        this.wireFormat = wireFormat;
        this.bindAddressURI = bindAddr.toString();
        closed = new SynchronizedBoolean(false);
        started = new SynchronizedBoolean(false);
        try {
            serverSocket = createServerSocket(bindAddr);
            log.info("Listening for connections at: " + bindAddr);
        }
        catch (Exception se) {
            System.out.println(se);
            se.printStackTrace();
            JMSException jmsEx = new JMSException("Bind to " + bindAddressURI + " failed: " + se.getMessage());
            jmsEx.setLinkedException(se);
            throw jmsEx;
        }
    }

    public TcpTransportServerChannel(WireFormat wireFormat, ServerSocket serverSocket) {
        this.wireFormat = wireFormat;
        this.serverSocket = serverSocket;
        closed = new SynchronizedBoolean(false);
        started = new SynchronizedBoolean(false);
        this.bindAddressURI = serverSocket.getInetAddress().toString();
        log.info("Listening for connections at: " + bindAddressURI);
    }

    /**
     * close the ServerChannel
     */
    public void stop() {
        if (closed.commit(false, true)) {
            super.stop();
            try {
                if (serverSocket != null) {
                    serverSocket.close();
                    serverSocketThread.join();
                    serverSocketThread = null;
                }
            }
            catch (Throwable e) {
                log.warn("Caught while closing: " + e + ". Now Closed", e);
            }
        }
    }

    /**
     * start listeneing for events
     *
     * @throws JMSException if an error occurs
     */
    public void start() throws JMSException {
        super.start();
        if (started.commit(false, true)) {
            serverSocketThread = new Thread(this, toString());
            serverSocketThread.setDaemon(true);
            serverSocketThread.start();
        }
    }

    /**
     * @return pretty print of this
     */
    public String toString() {
        return "TcpTransportServerChannel@" + bindAddressURI;
    }

    /**
     * pull Sockets from the ServerSocket
     */
    public void run() {
        while (!closed.get()) {
            Socket socket = null;
            try {
                socket = serverSocket.accept();
                if (socket != null) {
                    // have thread per channel for sending messages and a thread for receiving them
                    PooledExecutor executor = null;
                    if (useAsyncSend) {
                        executor = new PooledExecutor(new BoundedBuffer(maxOutstandingMessages), 1);
                    }
                    TcpTransportChannel channel = new TcpTransportChannel(wireFormat, socket, executor);
                    addClient(channel);
                }
            }
            catch (Exception e) {
                if (!closed.get()) {
                    log.warn("run()", e);
                }
            }
        }
    }

    // Properties
    //-------------------------------------------------------------------------
    public boolean isUseAsyncSend() {
        return useAsyncSend;
    }

    public void setUseAsyncSend(boolean useAsyncSend) {
        this.useAsyncSend = useAsyncSend;
    }

    public int getMaxOutstandingMessages() {
        return maxOutstandingMessages;
    }

    public void setMaxOutstandingMessages(int maxOutstandingMessages) {
        this.maxOutstandingMessages = maxOutstandingMessages;
    }

    public int getBacklog() {
        return backlog;
    }

    public void setBacklog(int backlog) {
        this.backlog = backlog;
    }

    // Implementation methods
    //-------------------------------------------------------------------------

    /**
     * Factory method to create a new ServerSocket
     *
     * @throws UnknownHostException
     * @throws IOException
     */
    protected ServerSocket createServerSocket(URI bind) throws UnknownHostException, IOException {
        ServerSocket answer = null;
        String host = bind.getHost();
        host = (host == null || host.length() == 0) ? "localhost" : host;
        InetAddress addr = InetAddress.getByName(host);
        if (addr.equals(InetAddress.getLocalHost())) {
            answer = new ServerSocket(bind.getPort(), backlog);
        }
        else {
            answer = new ServerSocket(bind.getPort(), backlog, addr);
        }
        return answer;
    }
}