/** 
 * 
 * Copyright 2004 Protique Ltd
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License. 
 * 
 **/
package org.codehaus.activemq;

import javax.jms.Destination;
import javax.jms.Message;
import javax.jms.MessageConsumer;
import javax.jms.MessageListener;
import javax.jms.MessageProducer;
import javax.jms.Session;
import javax.jms.TextMessage;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Date;
import java.util.Iterator;
import java.util.List;

/**
 * @version $Revision: 1.18 $
 */
public class JmsSendReceiveTestSupport extends TestSupport implements MessageListener {
    protected int messageCount = 100;
    protected String[] data;

    protected Session session;
    protected MessageConsumer consumer;
    protected MessageProducer producer;
    protected Destination destination;

    protected List messages = Collections.synchronizedList(new ArrayList());
    protected boolean topic = true;
    protected boolean durable = false;
    protected final Object lock = new Object();
    protected boolean verbose = false;

    public void testSendReceive() throws Exception {
        messages.clear();

        for (int i = 0; i < data.length; i++) {
            Message message = session.createTextMessage(data[i]);

            if (verbose) {
                System.out.println("About to send a message: " + message + " with text: " + data[i]);
            }

            producer.send(destination, message);
        }

        waitForMessagesToBeDelivered();

        int counter = 0;
        List copyOfMessages = Arrays.asList(messages.toArray());
        if (data.length != copyOfMessages.size()) {
            for (Iterator iter = copyOfMessages.iterator(); iter.hasNext();) {
                TextMessage message = (TextMessage) iter.next();
                System.out.println("<== " + counter++ + " = " + message);
            }
        }

        assertEquals("Not enough messages received", data.length, messages.size());

        for (int i = 0; i < data.length; i++) {
            TextMessage received = (TextMessage) messages.get(i);
            String text = received.getText();

            if (verbose) {
                System.out.println("Received Text: " + text);
            }

            assertEquals("Message: " + i, data[i], text);
        }
    }

    protected void waitForMessagesToBeDelivered() {
        // lets wait a little while - note we might miss some notify() calls
        // as we might not be always waiting for the lock when the notify occurs
        int count = 0;
        while (count < data.length) {
            waitForTimeOrNotify(4000L);

            // we may have missed a few notify() calls, so skip them
            int size = messages.size();
            if (size > count) {
                count = size;
            }
            else {
                count++;
            }
        }

        if (messages.size() < data.length) {
            // one more just for good luck
            waitForTimeOrNotify(5000L);
        }
    }

    private void waitForTimeOrNotify(long timeout) {
        try {
            synchronized (lock) {
                lock.wait(timeout);
            }
        }
        catch (InterruptedException e) {
            System.out.println("Caught: " + e);
        }
    }


    protected void setUp() throws Exception {
        super.setUp();
        String temp = System.getProperty("messageCount");
        if (temp != null) {
            int i = Integer.parseInt(temp);
            if (i > 0) {
                messageCount = i;
            }
        }
        System.out.println("Message count for test case is: " + messageCount);

        data = new String[messageCount];
        for (int i = 0; i < messageCount; i++) {
            data[i] = "Text for message: " + i + " at " + new Date();
        }
    }

    public synchronized void onMessage(Message message) {
        if (verbose) {
            System.out.println("Received message: " + message);
        }

        messages.add(message);
        synchronized (lock) {
            lock.notifyAll();
        }
    }
}
