/** 
 * 
 * Copyright 2004 Protique Ltd
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License. 
 * 
 **/
package org.codehaus.activemq.config;

import junit.framework.TestCase;
import org.codehaus.activemq.broker.Broker;
import org.codehaus.activemq.broker.BrokerContainer;
import org.codehaus.activemq.broker.impl.BrokerConnectorImpl;
import org.codehaus.activemq.spring.ActiveMQBeanDefinitionReader;
import org.codehaus.activemq.spring.ActiveMQBeanFactory;
import org.codehaus.activemq.spring.ActiveMQDtdResolver;
import org.codehaus.activemq.spring.SpringBrokerContainerFactory;
import org.codehaus.activemq.transport.TransportServerChannel;
import org.codehaus.activemq.transport.tcp.TcpTransportServerChannel;
import org.codehaus.activemq.transport.vm.VmTransportServerChannel;
import org.springframework.beans.factory.xml.XmlBeanFactory;
import org.springframework.core.io.ClassPathResource;
import org.springframework.core.io.FileSystemResource;
import org.springframework.core.io.Resource;
import org.w3c.dom.Document;

import javax.jms.JMSException;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.Transformer;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;
import java.io.File;
import java.util.List;

/**
 * @version $Revision: 1.9 $
 */
public class ConfigTest extends TestCase {
    public void testConfig() {
        ActiveMQBeanFactory factory = new ActiveMQBeanFactory("Cheese", new ClassPathResource("org/codehaus/activemq/config/example.xml"));

        Object value = factory.getBean("broker");

        assertTrue("Should have created a broker!", value != null);
        assertTrue("Should be a broker container: " + value, value instanceof BrokerContainer);

        BrokerContainer container = (BrokerContainer) value;
        Broker broker = container.getBroker();
        assertTrue("Should have a broker!", broker != null);

        assertEquals("Broker name not set!", "localhost", broker.getBrokerName());

        Object transport = factory.getBean("transport");
        assertTrue("Made transport", transport != null);

        List connectors = container.getConnectors();
        assertEquals("Should have created more connectors", 3, connectors.size());

        BrokerConnectorImpl connector1 = (BrokerConnectorImpl) connectors.get(0);
        TransportServerChannel serverChannel1 = connector1.getServerChannel();
        assertTrue(serverChannel1 instanceof VmTransportServerChannel);

        BrokerConnectorImpl connector2 = (BrokerConnectorImpl) connectors.get(1);
        TransportServerChannel serverChannel2 = connector2.getServerChannel();
        assertTrue(serverChannel2 instanceof TcpTransportServerChannel);
        TcpTransportServerChannel tcpChannel2 = (TcpTransportServerChannel) serverChannel2;
        assertEquals("backlog", 1000, tcpChannel2.getBacklog());
        assertEquals("maxOutstandingMessages", 50, tcpChannel2.getMaxOutstandingMessages());
        assertTrue("useAsyncSend", tcpChannel2.isUseAsyncSend());
    }

    public void testBerkeleyDBConfig() throws JMSException {
        createBroker(new FileSystemResource("src/sample-conf/bdb-example.xml"));

        File file = new File("target/berkeleydb");
        assertTrue("Created the file for the persistent store: " + file, file.exists());
    }


    public void testHowlConfig() throws JMSException {
        createBroker(new FileSystemResource("src/sample-conf/howl-example.xml"));

        File file = new File("target/howlTest/berkeleydb");
        assertTrue("Created the file for the persistent store: " + file, file.exists());
    }

    public void testJdbmConfig() throws JMSException {
        createBroker(new FileSystemResource("src/sample-conf/jdbm-example.xml"));

        File file = new File("target/XmlTest/jdbm");
        assertTrue("Created the file for the persistent store: " + file, file.exists());
    }

    public void testVmConfig() throws JMSException {
        createBroker(new FileSystemResource("src/sample-conf/vm-example.xml"));
    }

    public void testTransform() throws Exception {
        ClassPathResource resource = new ClassPathResource("org/codehaus/activemq/activemq-to-spring.xsl");
        StreamSource source = new StreamSource(resource.getInputStream(), resource.getURL().toString());
        Transformer transformer = ActiveMQBeanDefinitionReader.createTransformer(source);

        DocumentBuilder builder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
        builder.setEntityResolver(new ActiveMQDtdResolver());
        Document document = builder.parse(new ClassPathResource("org/codehaus/activemq/config/example.xml").getFile());

        transformer.transform(new DOMSource(document), new StreamResult(new File("target/example-spring.xml")));
    }

    public void testSpring() throws Exception {
        XmlBeanFactory factory = new XmlBeanFactory(new ClassPathResource("org/codehaus/activemq/config/spring-test.xml"));

        Object transport = factory.getBean("transport");
        assertTrue("Made transport", transport != null);

        System.out.println("Created transport: " + transport);
    }


    protected BrokerContainer createBroker(Resource resource) throws JMSException {
        String brokerName = getName();
        SpringBrokerContainerFactory factory = new SpringBrokerContainerFactory();
        factory.setResource(resource);
        BrokerContainer container = factory.createBrokerContainer(brokerName);

        assertTrue("Should have a broker container!", container != null);

        Broker broker = container.getBroker();
        assertTrue("Should have a broker!", broker != null);

        assertEquals("Broker name not set!", brokerName, broker.getBrokerName());

        container.start();
        container.stop();
        return container;

    }
}
